/*
 * channel.c:	Get and forward AC3 stream from DVB card to the
 *		S/P-DIF interface of sound cards supported by ALSA.
 *		(NO decoding!)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2002,2003 Werner Fink, <werner@suse.de>
 *
 * Changes and Additional Code:
 *
 * Copyright (C) 2003 Sven Goethel, <sven@jausoft.com>
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include "types.h"
#include "channel.h"
#include "ac3.h"		// AC3 is the only available stream with LiveView 
#include "lpcm.h"		// AC3 is the only available stream with LiveView 

// --- cInStream : Get AC3 stream for redirecting it to  S/P-DIF of a sound card--------

const uint_32 cInStream::PS1magic = char2int(0x00,0x00,0x01,0xBD);
uint_8  * cInStream::tsdata;
cBounce * cInStream::bounce;

cInStream::cInStream( uint_16 aPid, bool aPidIsDD, spdif *dev, ctrl_t &up, cBounce *bPtr)
:cReceiver(0, -1, 1, aPid), spdifDev(dev), setup(up)
{
    flags = 0;
    apid = aPid;
    apidIsDD = aPidIsDD;
    bfound = paklen = ptsoff = ptsflag = 0;
    syncword = 0xffffffff;
    bounce = bPtr;
    bounce->bank(0);
    tsdata = setup.buf + TRANSFER_START;

    PrimaryDevice = cDevice::PrimaryDevice();

    debug("INSTREAM cstr .. apid =%d, apidIsDD=%d\n", apid, apidIsDD);
}

cInStream::~cInStream(void)
{
    debug("INSTREAM dstr .. apid =%d, apidIsDD=%d\n", apid, apidIsDD);
    Detach();
    Activate(false);
    bounce->flush();
    clear_flag(BOUNDARY);
    clear_flag(FAILED);
    debug("INSTREAM dstr .. X\n");
}

void cInStream::Activate(bool onoff)
{
    if (onoff) {
	if (*spdifDev)
	    goto out;		// An other has opened the S/P-DIF
	if (test_setup(RESET)) {
	    clear_flag(FAILED);
	    clear_setup(RESET);
	}
	if (test_flag(FAILED) || test_flag(ACTIVE))
	    goto out;
	set_flag(ACTIVE);
	set_setup(LIVE);
	Start();
	pthread_yield();
    } else {
        debug("INSTREAM de-activate .. 1\n");
	clear_flag(ACTIVE);
	while (test_flag(RUNNING)) {
            debug("INSTREAM de-activate .. 3\n");
	    clear_flag(ACTIVE);
	    bounce->flush();
	    bounce->signal();
	    usleep(1000);
	    if ( ! Active() ) break;  // S.O.S
	}
	clear_flag(RUNNING);
	Cancel(1);
	clear_setup(LIVE);
	stream = NULL;
        debug("INSTREAM de-activate .. X\n");
    }
out:
    return;
}

void cInStream::Action(void)
{
    set_flag(RUNNING);
    iec60958 *tmp;

    clear_setup(AUDIO);
    spdifDev->Clear();
    spdifDev->Close();

    if( apidIsDD )
    {
	    debug("cInStream::Action ac3 stream\n");
	    stream = &ac3;
	    stream->stream_type = IEC_AC3;
	    stream->Reset();
	    stream->track = 0;
	    clear_setup(AUDIO);
    } else {
	    debug("cInStream::Action MP2 stream\n");
    	    MP2Dec.init();
	    stream = &pcm;
	    stream->stream_type = IEC_PCMCVRTED_MP2;
	    stream->Reset();
	    stream->track = 0;
	    stream->isDVD = true;
	    pcm.sample_rate = 48000;
	    set_setup(AUDIO);
    }

    if (!stream || !spdifDev->Open(stream, setup)) {
	esyslog("INSTREAM: can't open spdif interface");
	set_flag(FAILED);
	goto out;
    }

    dsyslog("INSTREAM: Forwarding bitstream thread started (pid=%d)\n", getpid());
    debug("INSTREAM: Forwarding bitstream thread started (pid=%d)\n", getpid());
    realtime("INSTREAM: ");
    bounce->takeio();
    pts.clear();
    tmp = stream;
    while (test_flag(ACTIVE)) {
	ssize_t len;

	if (test_setup(MUTE)) {
    	    pts.clear();
	    spdifDev->Clear();
	}

	if (!spdifDev->Synchronize(bounce))	// Wait on spdifDev->Signal()
	    break;

	if (test_setup(MUTE)) {
	    bounce->flush();
	    clear_flag(BOUNDARY);
	    continue;
	}

	if ((len = bounce->fetch(tsdata, spdifDev->Available(TRANSFER_MEM))))
	{
	    if( apidIsDD )
	    {
	        ScanTSforPS1(tsdata, len);
            } else {
		ScanTSforMP2(tsdata, len);
	    } 
	}

    }
    dsyslog("INSTREAM: Forwarding bitstream thread ended (pid=%d)\n", getpid());
    debug("INSTREAM: Forwarding bitstream thread ended (pid=%d)\n", getpid());
    spdifDev->Close();
    debug("INSTREAM: Forwarding bitstream thread ended 0\n");
    tmp->Reset();
out:
    stream = NULL;
    bounce->leaveio();
    clear_flag(RUNNING);
    debug("INSTREAM: Forwarding bitstream thread ended X\n");
}

void cInStream::Receive(uchar *b, int cnt)
{
    uint_8 off = 4;
    uint_8 start = 0;
    static uint_16 skip = 0;

    if (!test_setup(ACTIVE)) {
	Activate(false);
	goto out;
    } else if (!test_flag(RUNNING))
	Activate(true);

    if (skip) {
	debug("cInStream::Receive(skip=%d)\n", skip);
	skip--;
	goto out;
    }

    if (test_setup(MUTE)) {
	clear_flag(PAYSTART);
	goto out;
    }

    if (!b || *b != 0x47) {
	esyslog("INSTREAM: have seen broken TS packet");
	goto out;
    }

    if (cnt != TS_SIZE) {
	esyslog("INSTREAM: have seen broken TS packet");
	goto out;
    }

    start = (b[1] & PAY_START);

    // Start engine only if PES frame starts
    if (!test_flag(PAYSTART)) {
	if (!start)
	    goto out;
	set_flag(PAYSTART);
    }

    if (apid != (((uint_16)(b[1]&0x1F)<<8)|b[2]))
	goto out;

#if 0
    if (b[1] & TS_ERROR)
	TSerr++;			// Skip?

    if ((b[3] ^ TScount) & CONT_CNT_MASK) {
	if (/* TScount != -1 && */ ((b[3] ^ (TScount + 1)) & CONT_CNT_MASK))
	    TScnterr++;			// Skip?
	TScount = (b[3] & CONT_CNT_MASK);
    }
#endif
    if (b[3] & ADAPT_FIELD) {
	off += (b[4] + 1);
	if (off > 187)
	    goto out;
    }

    if (!bounce->store(&b[off], TS_SIZE-off, (start != 0)))
	skip = 20;
out:
    return;
}

inline void cInStream::ScanTSforPS1(const uint_8 *buf, const int cnt)
{
    const uint_8 *const tail = buf + cnt;

resync:
    if (bfound < 9) {
	//
	// Scan for payload length and PTS, the switch is used here
	// to be able to start at any offset (controlled by bfound)
	//
	switch (bfound) {
	case 0 ... 3:
	    //
	    // Scan for start sequence of the Private Stream 1
	    //
	    while (syncword != PS1magic) {
		if (buf >= tail)
		    goto out;
		syncword = (syncword << 8) | *buf++;
	    }
	    bfound = 4;
	case 4:
	    if (buf >= tail)
		goto out;
	    paklen  = (uint_16)(*buf++) << 8;
	    bfound++;
	case 5:
	    if (buf >= tail)
		goto out;
	    paklen |= (*buf++);
	    paklen += 6;			// Full packet length
	    bfound++;
	case 6:
	    if (buf >= tail)
	    {
		goto out;
	    }
	    if (((*buf++) & 0xC0) != 0x80) {	// First mpeg2 PTS flag (PTS if Mpeg2)
		// Reset for next scan
		bfound = paklen = ptsoff = ptsflag = 0;
		syncword = 0xffffffff;
		clear_flag(BOUNDARY);
		esyslog("INSTREAM: broken TS stream");
		goto resync;			// Check for rest
	    }
	    if ((*buf) & 0x04)			// Is the sub stream aligned?
		set_flag(BOUNDARY);
	    bfound++;
	case 7:
	    if (buf >= tail)
		goto out;
            ptsflag = (*buf)  >> 6;
	    if ((*buf++) & 0xC0)		// Second mpeg2 PTS flag (PTS only)
		set_flag(BOUNDARY);
	    bfound++;
	case 8:
	    if (buf >= tail)
		goto out;
	    ptsoff = (*buf++);			// Third mpeg2 PTS flag (PTS offset)
	    bfound++;
	default:
	    break;
	}
    }

    //
    // Start only at a logical begin of a sub stream
    //
    if (!test_flag(BOUNDARY) || !test_flag(RUNNING) || !test_flag(ACTIVE)) {
	bfound = paklen = ptsoff = ptsflag = 0; // Reset for next scan
	syncword = 0xffffffff;
	clear_flag(BOUNDARY);
	goto out;
    }

    if( ptsflag && ptsoff>4 ) {
	uint64_t devSTC;
	if ( PrimaryDevice->GetSTC(devSTC) )
	     pts.process(buf, (uint_32)(devSTC/90UL), true, NULL);
    }

    //
    // Skip PTS
    //
    while (ptsoff) {
	register int skip;
	if ((skip = tail - buf) <= 0)
	    goto out;
	if (skip > ptsoff)
	    skip = ptsoff;
	buf    += skip;
	bfound += skip;
	ptsoff -= skip;
    }

    //
    // Send data to payload scanner
    //
    while (bfound < paklen) {
	register int transmit, pay;
	if ((transmit = tail - buf) <= 0)
	    goto out;
	if (transmit > (pay = paklen - bfound))	// The rest is handled below
	    transmit = pay;
	(void)spdifDev->Forward(buf, transmit, bounce);
	buf    += transmit;
	bfound += transmit;
    }

    //
    // Bytes found are equal to packet length if we have found the PES packet,
    // for the rest restart scanning.
    //
    if (paklen && (bfound == paklen)) {
	bfound = paklen = ptsoff = ptsflag = 0; // Reset for next scan
	syncword = 0xffffffff;
	clear_flag(BOUNDARY);
	if (buf >= tail)
	    goto out;
	goto resync;				// Check the rest for next frame
    }
out:
    return;
}


inline void cInStream::ScanTSforMP2(const uint_8 *tsdata, int len)
{
      const uint_8 * data = tsdata;
      int datalen         = len;

      static unsigned long Frame = 0;

      debug("cInStream::ScanTSforMP2 MP2 IN paklen %d, 0x%X, len %d, 0x%X!\n", 
		paklen, paklen, len, len);

      if (!test_flag(BOUNDARY)) {
	bfound = paklen = ptsoff = ptsflag = 0; // Reset for next scan
	debug("cInStream::ScanTSforMP2 MP2 IN reset: PTS, paklen\n");
      }

      if (!test_flag(ACTIVE) || !test_flag(RUNNING))
	return;

      while ( test_flag(ACTIVE) && test_flag(RUNNING) && len > 0 )
      {
        const uint_8 * pStart    = data;
        const uint_8 * pEnd      = data+len;

	// resync to payload ..
	if ( paklen == 0 ) {
		debug("cInStream::ScanTSforMP2 MP2 RESYNC pos 0x%X, paklen %d 0x%X, len %d 0x%X !\n", 
			pStart-tsdata, paklen, paklen, len, len);

                while ( test_flag(ACTIVE) && test_flag(RUNNING) && pEnd-pStart > 9 ) {
                    if ( pStart[0]==0x00 && pStart[1]==0x00 && pStart[2]==0x01 && 
			 0xC0 <= pStart[3] && pStart[3] <= 0xDF ) 
			break;
		    pStart++;
		}

		if ( pEnd-pStart <= 9 ) {
		        debug("cInStream::ScanTSforMP2 MP2 RESYNC FAILED pos 0x%X, paklen %d 0x%X, len %d 0x%X !\n", 
				pStart-tsdata, paklen, paklen, len, len);
			goto hardsync;
			return;
		}

	        if ( (pStart[6] & 0xC0) == 0x80) { // First mpeg2 PTS flag (PTS if Mpeg2)
		    set_flag(BOUNDARY);
		}

    		if (!test_flag(BOUNDARY))
		{
		        debug("cInStream::ScanTSforMP2 MP2 RESYNC FAILED no BOUNDARY\n");
			goto hardsync;
		}

		if ( stream->track == 0) {
			stream->track = (pStart[3] & 7) + 1;
			debug("cInStream::ScanTSforMP2 MP2 RESYNC set atrack 0x%X -> %d\n",
				(int)pStart[3], stream->track);
		} 

		paklen   = ((uint_16)(pStart[4]) << 8) + 
			   ((uint_16)(pStart[5])) + 
			   6 ; // cPStream::packetLength

		Frame++;

    		ptsflag = pStart[7] >> 6;
		if( ptsflag ) {
			uint64_t devSTC;
			if ( PrimaryDevice->GetSTC(devSTC) )
			     pts.process(&pStart[9], (uint_32)(devSTC/90UL), true, &MP2Dec);
		}

		datalen  = len;
		data     = pStart;

		datalen -= (int)  (pStart[8]);	// cPStream::PESHeaderLength
		data    += (int)  (pStart[8]); // cPStream::PESHeaderLength

		datalen -= 9; // skip mandatory header bytes
		data    += 9; // skip mandatory header bytes

		if ( pEnd-data <= 9 ) {
		        debug("cInStream::ScanTSforMP2 MP2 RESYNC FAILED (2) pos 0x%X, paklen %d 0x%X, len %d 0x%X !\n", 
				data-pStart, paklen, paklen, len, len);
			goto hardsync;
			return;
		}

		debug("cInStream::ScanTSforMP2 MP2 RESYNCED pos 0x%X, paklen %d 0x%X !\n", 
			data-pStart, paklen, paklen);
	}
	pts.done(); // notify time now: how much has been seen !

	debug("cInStream::ScanTSforMP2 MP2 data pos 0x%X\n", data-pStart);

	if ( paklen < len ) {
		datalen = paklen;
		MP2Dec.pushData ( data, datalen );
		data   += paklen;
		len    -= paklen;
		paklen = 0;
     		clear_flag(BOUNDARY);
		debug("cInStream::ScanTSforMP2 MP2 data WRAPPING datalen %d, 0x%X, paklen %d, 0x%X, len %d, 0x%X\n", 
			datalen, datalen, paklen, paklen, len, len);
	} else {
		paklen -= len;
		len        = 0;
		MP2Dec.pushData ( data, datalen );
	}


	while ( test_flag(ACTIVE) && test_flag(RUNNING) )
	{
	    debug("cInStream::ScanTSforMP2 MP2 decode datalen %d, 0x%X, paklen %d, 0x%X, len %d, 0x%X, delay=%lu\n", 
			datalen, datalen, paklen, paklen, len, len, MP2Dec.getDelay());
	    if ( ! MP2Dec.decode () )
		break;

	    if( MP2Dec.pcm_netto_size > 0 ) {
	        debug("cInStream::ScanTSforMP2 MP2 FORWARD: %d\n", MP2Dec.pcm_netto_size);
	        (void)spdifDev->Forward(
			(const uchar*)(MP2Dec.lpcmFrame().Data), 
			MP2Dec.pcm_netto_size,
			bounce);
            }
	}
     }

     return;

hardsync:
     debug ("cInStream::ScanTSforMP2 RESYNC HARDSYNC PTS\n");
     paklen = 0;
     clear_flag(BOUNDARY);
     clear_flag(PAYSTART);
     pts.clear();
     bounce->flush();
     return;
}

inline void cInStream::HardSyncReset()
{
     debug ("cInStream::HardSyncReset RESYNC HARDSYNCRESET PTS\n");
     paklen=0;
     MP2Dec.init();

     if(spdifDev) {
	     spdifDev->Clear();
	     spdifDev->Pause(true);
	     (void) spdifDev->Synchronize(bounce);
     }	

     if(bounce) {
	     bounce->flush();
     }

     clear_flag(BOUNDARY);
     clear_flag(PAYSTART);

     if(spdifDev) {
	     spdifDev->Pause(false);
     }
}

// --- cControlPerm : Extend Access to cPlayer from cControl ..

class cControlPerm : public cControl
{
private:
	cControlPerm(cPlayer *Player, bool Hidden = false)
	: cControl(Player, Hidden) 
	{ }
	virtual void Hide(void) {}

public:
	cPlayer *Player()
	{ return player; }
};

// --- cChannelOutSPDif : Live AC3 stream over S/P-DIF of a sound card with ALSA ---------
cBounce * cChannelOutSPDif::bounce;

cChannelOutSPDif::cChannelOutSPDif(spdif &dev, ctrl_t &up, cBounce * bPtr, const char *script)
: SPDIFmute(script), spdifDev(&dev), setup(up)
{
    in = NULL;
    Apid = 0x1FFF;
    ApidIsDD = false;
    ApidOrig = 0x1FFF;
    ApidLocked=false;
    replaying = false;
    replayControl=NULL;
    replayName=NULL;
    bounce = bPtr;
    Activate(true);
}

cChannelOutSPDif::~cChannelOutSPDif(void)
{
    Activate(false);
}

void cChannelOutSPDif::AudioSwitch(uint_16 apid, bool apidisdd)
{
    clear_flag(CHANNELSET);

    if ( !replaying ) {
	    if ( apid == ApidOrig || ApidOrig==0x1FFF || ApidOrig==0) 
		ApidLocked=false;
	    else
		ApidLocked=true;

	    if (in)
		    ActivateInStream(false);

	    Apid = apid;
	    ApidIsDD = apidisdd;

	    if(!ApidLocked)
		    ApidOrig = Apid;

            if (!test_setup(MP2ENABLE) && !apidisdd) {
		    debug("cChannelOutSPDif::AudioSwitch mp2 disabled .. apid =%d, apidisdd=%d\n", Apid, ApidIsDD);
		    goto out;
	    }
	    debug("cChannelOutSPDif::AudioSwitch .. apid =%d, apidisdd=%d\n", Apid, ApidIsDD);

	    // Valid pid for AC3 or MP2 to be attached too?
	    if ( Apid )
	    {
		ActivateInStream(true);
	    } else
		test_setup(RESET);
	    IfNeededMuteSPDIF();	// On close the S/P-DIF is not muted anymore

    } else if ( replayControl!=NULL && replaying) {
    	int numtracks = ((cControlPerm *)replayControl)->Player()->NumAudioTracks();
	int num       = (int)apid-1;
	if(0<=num && num<numtracks) {
	        debug("cChannelOutSPDif::AudioSwitch player .. num =%d\n", num);
		((cControlPerm *)replayControl)->Player()->SetAudioTrack(num);
	}
    }
out:
    set_flag(CHANNELSET);

    debug("cChannelOutSPDif::AudioSwitch X\n");
}

void cChannelOutSPDif::AudioSwitch()
{
    uint_16 apid;
    bool    apidisdd;

    if(ApidLocked) {
        debug("cChannelOutSPDif::AudioSwitch auto n.a. locked\n");
	return;
    }

    clear_flag(CHANNELSET);

    if ( !GetCurrentAudioTrack(apid, apidisdd))
	return;

    AudioSwitch(apid, apidisdd);
}

void cChannelOutSPDif::ChannelSwitch(const cDevice *Device, int channelNumber)
{
    cDevice *PrimaryDevice = cDevice::PrimaryDevice();
    cChannel *channel=NULL;

    // LiveView is only on the primary device possible, isn't it?
    if (!Device->IsPrimaryDevice())
	goto out;

    ApidLocked=false;
    ChannelNumber = channelNumber; 

    clear_flag(CHANNELSET);

    if (in)
	ActivateInStream(false);

    debug("cChannelOutSPDif::ChannelSwitch num=%d, repl=%d\n", ChannelNumber, replaying);

    // If we're noticed only for channel switch or
    // a replay is running we're done
    if (!ChannelNumber || replaying)
	goto out;

    // No channel, no audio
    if (!(channel = Channels.GetByNumber(ChannelNumber)))
	goto out;

#if defined(VDRVERSNUM) && (VDRVERSNUM > 10130)
    // Sanity check not be attached in transfer mode
    // in this case the replay part should taken the data
    if (PrimaryDevice != cDevice::ActualDevice())
	goto out;

    // Next sanity check not be attached in transfer mode
    // from the primary device to the primary device.
    // As metioned above the replay part should taken the data
    if (PrimaryDevice == cTransferControl::ReceiverDevice())
	goto out;
#else
    if (!PrimaryDevice->ProvidesChannel(channel, Setup.PrimaryLimit))
	goto out;
#endif

    AudioSwitch();

    debug("cChannelOutSPDif::ChannelSwitch OK\n");
out:
    debug("cChannelOutSPDif::ChannelSwitch X\n");
    return;
}

void cChannelOutSPDif::Replaying(const cControl *Control, const char *Name)
{
    replayControl=Control;
    replayName=Name;

    replaying = (Name != NULL);

    debug("cChannelOutSPDif:: Replaying %d %s\n", replaying, Name==NULL?"n.a.":Name,
	Control==NULL?"no contrl":"has control");

    if (replaying) {
	ActivateInStream(false);
        clear_flag(CHANNELSET);
    } else {
	ActivateInStream(true);
        set_flag(CHANNELSET);
    }
}

void cChannelOutSPDif::IfNeededMuteSPDIF(void)
{
    if (SPDIFmute) {    // Normal TV mode
	char *cmd = NULL;
	asprintf(&cmd, "%s %s 2> /dev/null", SPDIFmute, test_setup(MUTE) ? "mute" : "unmute");
	if (cmd) {
	    system(cmd);
	    free(cmd);
	    cmd = NULL;
	}
    }
}

void cChannelOutSPDif::ActivateInStream(bool on)
{
    cDevice *PrimaryDevice = cDevice::PrimaryDevice();

    if (on) {
        debug("cChannelOutSPDif:: ActivateInStream true 1\n");
        if(in) {
		ActivateInStream(false);
        }
	in = new cInStream(Apid, ApidIsDD, spdifDev, setup, bounce);
	if (!in) {
	    esyslog("ERROR: out of memory");
	    debug("cChannelOutSPDif::AudioSwitch ERROR: out of memory\n");
	    Apid = 0x1FFF;
	} else {
	    PrimaryDevice->AttachReceiver(in);
	    debug("cChannelOutSPDif::AudioSwitch true attached\n");
	}
        debug("cChannelOutSPDif:: ActivateInStream true X\n");
    } else {
        debug("cChannelOutSPDif:: ActivateInStream false 1\n");
	if(in!=NULL) {
		delete in;
		in = NULL;
	}
	bounce->flush();
	Apid = 0x1FFF;
	debug("cChannelOutSPDif:: ActivateInStream false X\n");
    }
}

void cChannelOutSPDif::Activate(bool on)
{
    if (on) {
        debug("cChannelOutSPDif:: Activate true 1\n");
	if (!test_flag(ACTIVE))
	{
		set_flag(ACTIVE);
		Start();
		pthread_yield();
        }
        debug("cChannelOutSPDif:: Activate true X\n");
    } else {
        debug("cChannelOutSPDif:: Activate false 1\n");
	clear_flag(ACTIVE);
	while (test_flag(RUNNING)) {
	    clear_flag(ACTIVE);
	    usleep(1000);
	    if ( ! Active() ) break;  // S.O.S
	}
	clear_flag(RUNNING);
	Cancel(1);
        debug("cChannelOutSPDif:: Activate false X\n");
    }
}

void cChannelOutSPDif::Action(void)
{
    set_flag(RUNNING);

    dsyslog("cChannelOutSPDif::Action started\n");
    debug("cChannelOutSPDif::Action started\n");
    while (test_flag(ACTIVE)) {
	uint_16 t_apid;
        bool    t_apidIsDD;

	if ( test_flag(CHANNELSET) && in!=NULL && 
             GetCurrentAudioTrack(t_apid, t_apidIsDD) ) 
        {
		if ( ! ( t_apid==Apid && t_apidIsDD==ApidIsDD ) && t_apid!=0 && !ApidLocked) {
		    debug("cChannelOutSPDif::Action AudioSwitch stream has changed X!\n");
		    debug("\tapid %d (%d) -> %d(%d)\n", Apid, ApidIsDD, t_apid, t_apidIsDD);
		    AudioSwitch();
		}
	}

	usleep(200 * 1000); // 200ms -> 5 x per sec.

    }
    clear_flag(RUNNING);
    debug("cChannelOutSPDif::Action ended\n");
    dsyslog("cChannelOutSPDif::Action ended\n");
}

bool cChannelOutSPDif::GetCurrentAudioTrack(uint_16 &apid, bool &apidIsDD)
{
    if ( replaying ) return false;

    cChannel *channel=NULL;
    if (!(channel = Channels.GetByNumber(ChannelNumber)))
    {
        debug("cChannelOutSPDif::GetCurrentAudioTrack AudioSwitch cannot get channel !\n");
	return false;
    }
    return GetCurrentAudioTrack(apid, apidIsDD, channel);
}

bool cChannelOutSPDif::GetCurrentAudioTrack(uint_16 &apid, bool &apidIsDD, cChannel *channel)
{
    cDevice *PrimaryDevice = cDevice::PrimaryDevice();
    uint_16 apid_mp  = 0x1FFF;
    uint_16 apid_dd  = 0x1FFF;
    int CurrentAudioTrack;

    if(!channel) 
    {
        debug("cChannelOutSPDif::GetCurrentAudioTrack AudioSwitch given channel NULL!\n");
	return false;
    }

    (void) PrimaryDevice->GetAudioTracks(&CurrentAudioTrack);

    switch (CurrentAudioTrack) {
		case 0:
			apid_dd = channel->Dpid1();
			apid_mp = channel->Apid1();
			break;
		case 1:
			apid_dd = channel->Dpid2();
			apid_mp = channel->Apid2();
			break;
		default:
        		debug("cChannelOutSPDif::GetCurrentAudioTrack no audio track provided!\n");
			apid = 0;
    }

    apid = apid_dd;

    if ( apid_dd == 0x1FFF || apid_dd == 0 ) {
	apidIsDD = false;
        apid = apid_mp;
    } else {
	apidIsDD = true;
        apid = apid_dd;
    }
    
    if(apid >= 0x1FFF) 
    {
	apid = 0;
	debug("cChannelOutSPDif::GetCurrentAudioTrack no audio pid provided!\n");
    }

    return apid!=0;
}

bool cChannelOutSPDif::GetAudioTrack(int num, bool dd, int &apid)
{
    if ( !replaying ) {
	    cChannel *channel=NULL;
	    if (!(channel = Channels.GetByNumber(ChannelNumber)))
	    {
		debug("cChannelOutSPDif::GetAudioTrack cannot get channel !\n");
		return false;
	    }

	    switch (num) {
			case 0:
				apid = dd?channel->Dpid1():channel->Apid1();
				break;
			case 1:
				apid = dd?channel->Dpid2():channel->Apid2();
				break;
			default:
				debug("cChannelOutSPDif::GetAudioTrack invalid num !\n");
				apid = 0;
	    }

	    if(apid >= 0x1FFF) 
	    {
		apid = 0;
		debug("cChannelOutSPDif::GetCurrentAudioTrack no audio pid provided!\n");
	    }
    } else if ( replayControl!=NULL && replaying) {
    	int numtracks = ((cControlPerm *)replayControl)->Player()->NumAudioTracks();
	debug("cChannelOutSPDif::GetAudioTrack numtracks=%d, asked=%d dd=%d", numtracks, num, dd);
	if(0<=num && num<numtracks && !dd) {
		apid=num+1;
	} else {
		apid=0;
	}
	debug(" result=%d\n", apid);
    }

    return apid!=0;
}

uint_16  cChannelOutSPDif::getApid() 
{ 
	if ( in!=NULL && !replaying ) {
		return in->getApid(); 
        } else if ( replayControl!=NULL && replaying) {
		int apid;
		(void) ((cControlPerm *)replayControl)->Player()->GetAudioTracks(&apid);
		debug("cChannelOutSPDif::getApid players current %d\n", apid);
		return (uint_16) apid+1;
	}
	return Apid; 
}

bool cChannelOutSPDif::getApidIsDD() 
{ 
	if ( in!=NULL && !replaying ) 
		return in->getApidIsDD(); 
	else  if ( replaying ) 
	{
		debug("cChannelOutSPDif::getApidIsDD player - unknown\n");
		return false;
	}
	return ApidIsDD; 
}


