/*
 * pts.h:	P(resentation) T(ime) S(tamps) handling
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2003 Werner Fink, <werner@suse.de>
 *
 * Changes and Additional Code:
 *
 * Copyright (C) 2003 Sven Goethel, <sven@jausoft.com>
 */

#ifndef _PTS_H
#define _PTS_H

#include <unistd.h>
#include <stdint.h>
#include <sys/time.h>
#include "types.h"
#include "any_decoder.h"

# define debug_pts(args...)
// # define debug_pts(args...)	fprintf(stderr, args)

enum ePTSstate { eOK = 0, eMISSED = -1, eDOSKIP = 1 };

class cPTS {
private:
    // to hold the current state
    uint_32 pts, now;
    uint_32 start_stc;
    uint_32 start_pts;
    struct timeval ts;

public:
    static inline uint_32 PTSinMS(const uchar *pts)
    {
	return (uint_32)
	((   ((((uint64_t)pts[0]) & 0x0E) << 29)
	   | (( (uint64_t)pts[1])         << 22)
	   | ((((uint64_t)pts[2]) & 0xFE) << 14)
	   | (( (uint64_t)pts[3])         <<  7)
	   | ((((uint64_t)pts[4]) & 0xFE) >>  1))/90);
    }

/*

    sender (VDR)                    <->          receiver (spdif)
        |                                            |
     forward credits (ms)            ->           use credits if not expired
   + current time stamp for expire                if no data for credits repeat last frame

                                    <-            if expired put backward credits
                                                  of repeated frames

     if expired get backward credits
     for skipping                    ->           do skipping 
 */

    cPTS(void) 
    : pts(0), now(0), 
      start_stc(0),
      start_pts(0)
    { };

    inline void mark(const uchar *ptsdata)
    {
	pts = PTSinMS(ptsdata);
    };
    inline ePTSstate compare(const uchar *ptsdata, const uint_32 tolerance)
    {
	mark(ptsdata);
	bool lower = (now - tolerance <= pts);
	bool upper = (pts <= now + tolerance);
	ePTSstate ret;

	if (upper && lower)
	    ret = eOK;			// OK
	else if (!upper)
	    ret = eMISSED;		// MISSED frames
	else //  !lower
	    ret = eDOSKIP;		// DOSKIP frames
	return ret;
    };
    inline void clear()
    {
	pts = 0;
	reset();
    }
    inline void reset()
    {
	now = pts;
	gettimeofday(&ts, 0);
    };
    inline void done()
    {
	struct timeval cur, diff;
	gettimeofday(&cur, 0);
	timersub(&cur, &ts, &diff);
	now += (diff.tv_sec*1000+diff.tv_usec/1000);
	gettimeofday(&ts, 0);
    };
    inline uint_32 delay()
    {
	return now - pts;
    };
    inline uint_32 getNow() { return now; }

    inline uint_32 getPTS() { return pts; }

    
    inline long process( const uchar *ptsdata, uint_32 stc, bool dosleep, AnyDecoder * AnyDecOptional )
    {
	long    delay_pts=0;

	if ( ptsdata==NULL) 
		goto out;

	if ( stc == 0 ) 
		goto out;

	// if cleared -> setup
	if ( pts==0 ) {
		mark(ptsdata); //  pts setup
		reset();	    // pts -> now
		start_pts=pts;
		start_stc=stc;
		debug_pts("PTS mark reset stc, pts: %4lums - %4lums\n", stc, pts);
	} else {
		done(); // notify time now: how much has been seen !
		mark(ptsdata);
	}

	if ( AnyDecOptional != NULL )
		AnyDecOptional->setDelay(0);

	if ( stc > pts ) {
		delay_pts = stc - pts;
	} else {
		delay_pts = (pts-stc)*-1;
	}

	debug_pts("PTSDIFF stc-pts: %4lums - %4lums = %4ldms, ", stc, pts, delay_pts);

	if(abs(delay_pts)>5000) {
		delay_pts=0;
	}

	if ( delay_pts > 0 ) {
		if ( AnyDecOptional != NULL ) {
			AnyDecOptional->setDelay((uint_32)delay_pts);
			debug_pts(" mp2dec delay  %4ldms", delay_pts);
		}
	}  
	else if ( delay_pts < 0 && dosleep ) {
		delay_pts/=-2;

		done();
		uint_32 l_now=now;

		usleep ( (uint_32)delay_pts * 1000 );
		done();
		debug_pts(" slept %4lums", now-l_now);
		(void) l_now;
	}
	debug_pts("\n");

      out:
	return delay_pts;
    }
};

#endif

